"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TsCompiler = void 0;
const bs_logger_1 = require("bs-logger");
const path_1 = require("path");
const fs_1 = require("fs");
const lodash_memoize_1 = __importDefault(require("lodash.memoize"));
const mkdirp_1 = __importDefault(require("mkdirp"));
const compiler_utils_1 = require("./compiler-utils");
const constants_1 = require("../constants");
const logger_1 = require("../utils/logger");
const json_1 = require("../utils/json");
const messages_1 = require("../utils/messages");
const sha1_1 = require("../utils/sha1");
class TsCompiler {
    constructor(configSet) {
        this.configSet = configSet;
        this._memoryHost = {
            files: new Map(),
            resolvedModules: new Map(),
        };
        this._diagnosedFiles = [];
        this._projectVersion = 1;
        this._ts = configSet.compilerModule;
        this._logger = logger_1.rootLogger.child({ namespace: 'ts-compiler' });
        this._parsedTsConfig = this.configSet.parsedTsConfig;
        if (!this.configSet.isolatedModules) {
            this._createLanguageService();
        }
    }
    _createLanguageService() {
        const cacheDir = this.configSet.tsCacheDir;
        const serviceHostTraceCtx = {
            namespace: 'ts:serviceHost',
            call: null,
            [bs_logger_1.LogContexts.logLevel]: bs_logger_1.LogLevels.trace,
        };
        if (cacheDir) {
            mkdirp_1.default.sync(cacheDir);
            this._tsResolvedModulesCachePath = path_1.join(cacheDir, sha1_1.sha1('ts-jest-resolved-modules', '\x00'));
            try {
                const cachedTSResolvedModules = fs_1.readFileSync(this._tsResolvedModulesCachePath, 'utf-8');
                this._memoryHost.resolvedModules = new Map(json_1.parse(cachedTSResolvedModules));
            }
            catch (e) { }
        }
        this._parsedTsConfig.fileNames
            .filter((fileName) => !this.configSet.isTestFile(fileName))
            .forEach((fileName) => {
            this._memoryHost.files.set(fileName, {
                version: 0,
            });
        });
        this._cachedReadFile = this._logger.wrap(serviceHostTraceCtx, 'readFile', lodash_memoize_1.default(this._ts.sys.readFile));
        const moduleResolutionHost = {
            fileExists: lodash_memoize_1.default(this._ts.sys.fileExists),
            readFile: this._cachedReadFile,
            directoryExists: lodash_memoize_1.default(this._ts.sys.directoryExists),
            getCurrentDirectory: () => this.configSet.cwd,
            realpath: this._ts.sys.realpath && lodash_memoize_1.default(this._ts.sys.realpath),
            getDirectories: lodash_memoize_1.default(this._ts.sys.getDirectories),
        };
        const moduleResolutionCache = this._ts.createModuleResolutionCache(this.configSet.cwd, (x) => x, this._parsedTsConfig.options);
        const serviceHost = {
            getProjectVersion: () => String(this._projectVersion),
            getScriptFileNames: () => [...this._memoryHost.files.keys()],
            getScriptVersion: (fileName) => {
                var _a;
                const normalizedFileName = path_1.normalize(fileName);
                const version = (_a = this._memoryHost.files.get(normalizedFileName)) === null || _a === void 0 ? void 0 : _a.version;
                return version === undefined ? undefined : String(version);
            },
            getScriptSnapshot: (fileName) => {
                var _a;
                const normalizedFileName = path_1.normalize(fileName);
                const hit = this._isFileInCache(normalizedFileName);
                this._logger.trace({ normalizedFileName, cacheHit: hit }, 'getScriptSnapshot():', 'cache', hit ? 'hit' : 'miss');
                if (!hit) {
                    this._memoryHost.files.set(normalizedFileName, {
                        text: this._cachedReadFile(normalizedFileName),
                        version: 1,
                    });
                }
                const contents = (_a = this._memoryHost.files.get(normalizedFileName)) === null || _a === void 0 ? void 0 : _a.text;
                if (contents === undefined)
                    return;
                return this._ts.ScriptSnapshot.fromString(contents);
            },
            fileExists: lodash_memoize_1.default(this._ts.sys.fileExists),
            readFile: this._cachedReadFile,
            readDirectory: lodash_memoize_1.default(this._ts.sys.readDirectory),
            getDirectories: lodash_memoize_1.default(this._ts.sys.getDirectories),
            directoryExists: lodash_memoize_1.default(this._ts.sys.directoryExists),
            realpath: this._ts.sys.realpath && lodash_memoize_1.default(this._ts.sys.realpath),
            getNewLine: () => constants_1.LINE_FEED,
            getCurrentDirectory: () => this.configSet.cwd,
            getCompilationSettings: () => this._parsedTsConfig.options,
            getDefaultLibFileName: () => this._ts.getDefaultLibFilePath(this._parsedTsConfig.options),
            getCustomTransformers: () => this.configSet.customTransformers,
            resolveModuleNames: (moduleNames, containingFile) => {
                var _a;
                const normalizedContainingFile = path_1.normalize(containingFile);
                const currentResolvedModules = (_a = this._memoryHost.resolvedModules.get(normalizedContainingFile)) !== null && _a !== void 0 ? _a : [];
                return moduleNames.map((moduleName) => {
                    const resolveModuleName = this._ts.resolveModuleName(moduleName, containingFile, this._parsedTsConfig.options, moduleResolutionHost, moduleResolutionCache);
                    const resolvedModule = resolveModuleName.resolvedModule;
                    if (this.configSet.isTestFile(normalizedContainingFile) && resolvedModule) {
                        const normalizedResolvedFileName = path_1.normalize(resolvedModule.resolvedFileName);
                        if (!currentResolvedModules.includes(normalizedResolvedFileName)) {
                            currentResolvedModules.push(normalizedResolvedFileName);
                            this._memoryHost.resolvedModules.set(normalizedContainingFile, currentResolvedModules);
                        }
                    }
                    return resolvedModule;
                });
            },
        };
        this._logger.debug('created language service');
        this._languageService = this._ts.createLanguageService(serviceHost, this._ts.createDocumentRegistry());
    }
    getCompiledOutput(fileContent, fileName) {
        var _a;
        if (this._languageService) {
            this._logger.debug({ fileName }, 'getCompiledOutput(): compiling using language service');
            this._updateMemoryCache(fileContent, fileName);
            const output = this._languageService.getEmitOutput(fileName);
            if (this._tsResolvedModulesCachePath) {
                void (() => __awaiter(this, void 0, void 0, function* () {
                    yield fs_1.writeFile(this._tsResolvedModulesCachePath, json_1.stringify([...this._memoryHost.resolvedModules]), () => { });
                }))();
            }
            if (!this._diagnosedFiles.includes(fileName)) {
                this._logger.debug({ fileName }, 'getCompiledOutput(): computing diagnostics using language service');
                this._doTypeChecking(fileName);
            }
            if (!this.configSet.isTestFile(fileName)) {
                for (const [testFileName, resolvedModules] of this._memoryHost.resolvedModules.entries()) {
                    if (resolvedModules.includes(fileName) &&
                        !this._diagnosedFiles.includes(testFileName) &&
                        fs_1.existsSync(testFileName)) {
                        const testFileContent = (_a = this._memoryHost.files.get(testFileName)) === null || _a === void 0 ? void 0 : _a.text;
                        if (!testFileContent) {
                            this._updateMemoryCache(this._cachedReadFile(testFileName), testFileName);
                        }
                        this._logger.debug({ testFileName }, 'getCompiledOutput(): computing diagnostics using language service for test file which uses the module');
                        this._doTypeChecking(testFileName);
                    }
                }
            }
            if (output.emitSkipped) {
                throw new TypeError(`${path_1.relative(this.configSet.cwd, fileName)}: Emit skipped for language service`);
            }
            if (!output.outputFiles.length) {
                throw new TypeError(messages_1.interpolate("Unable to require `.d.ts` file for file: {{file}}.\nThis is usually the result of a faulty configuration or import. Make sure there is a `.js`, `.json` or another executable extension available alongside `{{file}}`.", {
                    file: path_1.basename(fileName),
                }));
            }
            return compiler_utils_1.updateOutput(output.outputFiles[1].text, fileName, output.outputFiles[0].text);
        }
        else {
            this._logger.debug({ fileName }, 'getCompiledOutput(): compiling as isolated module');
            const result = this._ts.transpileModule(fileContent, {
                fileName,
                transformers: this.configSet.customTransformers,
                compilerOptions: this._parsedTsConfig.options,
                reportDiagnostics: this.configSet.shouldReportDiagnostics(fileName),
            });
            if (result.diagnostics && this.configSet.shouldReportDiagnostics(fileName)) {
                this.configSet.raiseDiagnostics(result.diagnostics, fileName, this._logger);
            }
            return compiler_utils_1.updateOutput(result.outputText, fileName, result.sourceMapText);
        }
    }
    _isFileInCache(fileName) {
        return this._memoryHost.files.has(fileName) && this._memoryHost.files.get(fileName).version !== 0;
    }
    _updateMemoryCache(contents, fileName) {
        this._logger.debug({ fileName }, 'updateMemoryCache: update memory cache for language service');
        let shouldIncrementProjectVersion = false;
        const hit = this._isFileInCache(fileName);
        if (!hit) {
            this._memoryHost.files.set(fileName, {
                text: contents,
                version: 1,
            });
            shouldIncrementProjectVersion = true;
        }
        else {
            const cachedFileName = this._memoryHost.files.get(fileName);
            const previousContents = cachedFileName.text;
            if (previousContents !== contents) {
                this._memoryHost.files.set(fileName, {
                    text: contents,
                    version: cachedFileName.version + 1,
                });
                if (hit)
                    shouldIncrementProjectVersion = true;
            }
            if (!this._parsedTsConfig.fileNames.includes(fileName)) {
                shouldIncrementProjectVersion = true;
            }
        }
        if (shouldIncrementProjectVersion)
            this._projectVersion++;
    }
    _doTypeChecking(fileName) {
        if (this.configSet.shouldReportDiagnostics(fileName)) {
            const diagnostics = this._languageService.getSemanticDiagnostics(fileName).concat(this._languageService.getSyntacticDiagnostics(fileName));
            this._diagnosedFiles.push(fileName);
            this.configSet.raiseDiagnostics(diagnostics, fileName, this._logger);
        }
    }
}
exports.TsCompiler = TsCompiler;
