"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TsJestTransformer = void 0;
const config_set_1 = require("./config/config-set");
const constants_1 = require("./constants");
const json_1 = require("./utils/json");
const jsonable_value_1 = require("./utils/jsonable-value");
const logger_1 = require("./utils/logger");
const messages_1 = require("./utils/messages");
const ts_jest_compiler_1 = require("./compiler/ts-jest-compiler");
const sha1_1 = require("./utils/sha1");
class TsJestTransformer {
    constructor() {
        this.logger = logger_1.rootLogger.child({ namespace: 'ts-jest-transformer' });
        this.logger.debug('created new transformer');
    }
    configsFor(jestConfig) {
        const ccs = TsJestTransformer._cachedConfigSets.find((cs) => cs.jestConfig.value === jestConfig);
        let configSet;
        if (ccs) {
            this._transformCfgStr = ccs.transformerCfgStr;
            configSet = ccs.configSet;
        }
        else {
            const serializedJestCfg = json_1.stringify(jestConfig);
            const serializedCcs = TsJestTransformer._cachedConfigSets.find((cs) => cs.jestConfig.serialized === serializedJestCfg);
            if (serializedCcs) {
                serializedCcs.jestConfig.value = jestConfig;
                this._transformCfgStr = serializedCcs.transformerCfgStr;
                configSet = serializedCcs.configSet;
            }
            else {
                this.logger.info('no matching config-set found, creating a new one');
                configSet = new config_set_1.ConfigSet(jestConfig);
                const jest = Object.assign({}, jestConfig);
                const globals = (jest.globals = Object.assign({}, jest.globals));
                jest.name = undefined;
                jest.cacheDirectory = undefined;
                delete globals['ts-jest'];
                this._transformCfgStr = new jsonable_value_1.JsonableValue(Object.assign(Object.assign({ digest: configSet.tsJestDigest, babel: configSet.babelConfig }, jest), { tsconfig: {
                        options: configSet.parsedTsConfig.options,
                        raw: configSet.parsedTsConfig.raw,
                    } })).serialized;
                TsJestTransformer._cachedConfigSets.push({
                    jestConfig: new jsonable_value_1.JsonableValue(jestConfig),
                    configSet,
                    transformerCfgStr: this._transformCfgStr,
                });
            }
        }
        return configSet;
    }
    process(fileContent, filePath, jestConfig, transformOptions) {
        this.logger.debug({ fileName: filePath, transformOptions }, 'processing', filePath);
        let result;
        const configs = this.configsFor(jestConfig);
        const { hooks } = configs;
        const shouldStringifyContent = configs.shouldStringifyContent(filePath);
        const babelJest = shouldStringifyContent ? undefined : configs.babelJestTransformer;
        const isDefinitionFile = filePath.endsWith(constants_1.DECLARATION_TYPE_EXT);
        const isJsFile = constants_1.JS_JSX_REGEX.test(filePath);
        const isTsFile = !isDefinitionFile && constants_1.TS_TSX_REGEX.test(filePath);
        if (shouldStringifyContent) {
            result = `module.exports=${json_1.stringify(fileContent)}`;
        }
        else if (isDefinitionFile) {
            result = '';
        }
        else if (!configs.parsedTsConfig.options.allowJs && isJsFile) {
            this.logger.warn({ fileName: filePath }, messages_1.interpolate("Got a `.js` file to compile while `allowJs` option is not set to `true` (file: {{path}}). To fix this:\n  - if you want TypeScript to process JS files, set `allowJs` to `true` in your TypeScript config (usually tsconfig.json)\n  - if you do not want TypeScript to process your `.js` files, in your Jest config change the `transform` key which value is `ts-jest` so that it does not match `.js` files anymore", { path: filePath }));
            result = fileContent;
        }
        else if (isJsFile || isTsFile) {
            if (!TsJestTransformer._compiler) {
                TsJestTransformer._compiler = new ts_jest_compiler_1.TsJestCompiler(configs);
            }
            result = TsJestTransformer._compiler.getCompiledOutput(fileContent, filePath);
        }
        else {
            const message = babelJest ? "Got a unknown file type to compile (file: {{path}}). To fix this, in your Jest config change the `transform` key which value is `ts-jest` so that it does not match this kind of files anymore. If you still want Babel to process it, add another entry to the `transform` option with value `babel-jest` which key matches this type of files." : "Got a unknown file type to compile (file: {{path}}). To fix this, in your Jest config change the `transform` key which value is `ts-jest` so that it does not match this kind of files anymore.";
            this.logger.warn({ fileName: filePath }, messages_1.interpolate(message, { path: filePath }));
            result = fileContent;
        }
        if (babelJest) {
            this.logger.debug({ fileName: filePath }, 'calling babel-jest processor');
            result = babelJest.process(result, filePath, jestConfig, Object.assign(Object.assign({}, transformOptions), { instrument: false }));
        }
        if (hooks.afterProcess) {
            this.logger.debug({ fileName: filePath, hookName: 'afterProcess' }, 'calling afterProcess hook');
            const newResult = hooks.afterProcess([fileContent, filePath, jestConfig, transformOptions], result);
            if (newResult !== undefined) {
                return newResult;
            }
        }
        return result;
    }
    getCacheKey(fileContent, filePath, _jestConfigStr, transformOptions) {
        const configs = this.configsFor(transformOptions.config);
        this.logger.debug({ fileName: filePath, transformOptions }, 'computing cache key for', filePath);
        const { instrument = false, rootDir = configs.rootDir } = transformOptions;
        return sha1_1.sha1(this._transformCfgStr, '\x00', rootDir, '\x00', `instrument:${instrument ? 'on' : 'off'}`, '\x00', fileContent, '\x00', filePath);
    }
}
exports.TsJestTransformer = TsJestTransformer;
TsJestTransformer._cachedConfigSets = [];
