"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.VersionCheckers = void 0;
const semver_1 = require("semver");
const get_package_version_1 = require("./get-package-version");
const logger_1 = require("./logger");
const messages_1 = require("./messages");
const logger = logger_1.rootLogger.child({ namespace: 'versions' });
exports.VersionCheckers = {
    jest: createVersionChecker('jest', ">=26 <27"),
    typescript: createVersionChecker('typescript', ">=3.8 <5"),
    babelJest: createVersionChecker('babel-jest', ">=26 <27"),
    babelCore: createVersionChecker('@babel/core', ">=7.0.0-beta.0 <8"),
};
function checkVersion(name, expectedRange, action = 'warn') {
    let success = true;
    if (!('TS_JEST_DISABLE_VER_CHECKER' in process.env)) {
        const version = get_package_version_1.getPackageVersion(name);
        success = !!version && semver_1.satisfies(version, expectedRange);
        logger.debug({
            actualVersion: version,
            expectedVersion: expectedRange,
        }, 'checking version of %s: %s', name, success ? 'OK' : 'NOT OK');
        if (!action || success)
            return success;
        const message = messages_1.interpolate(version ? "Version {{actualVersion}} of {{module}} installed has not been tested with ts-jest. If you're experiencing issues, consider using a supported version ({{expectedVersion}}). Please do not report issues in ts-jest if you are using unsupported versions." : "Module {{module}} is not installed. If you're experiencing issues, consider installing a supported version ({{expectedVersion}}).", {
            module: name,
            actualVersion: version || '??',
            expectedVersion: rangeToHumanString(expectedRange),
        });
        if (action === 'warn') {
            logger.warn(message);
        }
        else if (action === 'throw') {
            logger.fatal(message);
            throw new RangeError(message);
        }
    }
    return success;
}
function rangeToHumanString(versionRange) {
    return new semver_1.Range(versionRange).toString();
}
function createVersionChecker(moduleName, expectedVersion) {
    let memo;
    const warn = () => {
        if (memo !== undefined)
            return memo;
        return (memo = checkVersion(moduleName, expectedVersion, 'warn'));
    };
    const raise = () => checkVersion(moduleName, expectedVersion, 'throw');
    return {
        raise,
        warn,
        forget() {
            memo = undefined;
        },
    };
}
